using OpenTK.Mathematics;
using OpenTK.Windowing.Common;
using OpenTK.Windowing.Desktop;

namespace GraphicsManager.Interfaces;

public interface IWindow : IParent
{
    public IRenderObject? HoveringControl { get; set; }
    public IGLFWGraphicsContext Context { get; }
    
    public bool ShowMissingChar { get; }
    public Matrix4 WindowSizeMatrix { get; }
    public Vector2i CS { get; }
    
    public bool LastFrameIsResize { get; }
    public bool UseSubFrames { get; set; }
    
    public bool UseLiveView { get; set; }

    public void CheckParent(IParent p, IRenderObject c, int xx, Vector2i di);
    public void CheckParent(IParent p);
    
    public event Action<MouseButtonEventArgs> MouseDown;
    
    public event Action<FileDropEventArgs> FileDrop;
    
    public event Action<MouseWheelEventArgs> MouseWheel;
    public event Action<MouseButtonEventArgs> MouseUp;

    public event Action<MouseMoveEventArgs> MouseMove;
    
    public event Action<KeyboardKeyEventArgs> KeyDown;
    
    public event Action<TextInputEventArgs> TextInput;
    public Vector2i Location { get; set; }
    
    public IRenderObject? focused { get; set; }
    
    public string ClipboardString { get; set; }
    
    public string Title { get; set; }
    
    public Vector2 MousePosition { get; set; }
    
    public bool CanControleUpdate { get; }

    void Invoke(Action A);
    
    bool InvokeRequired { get; }
    
    public bool LogFrames { get; set; }
    public int SubFrameCount { get; }
    public int Frame { get; }
    public void CopyFromFrontToBack(int x, int y, int width, int height, int NewX, int NewY);
    public void CopyFromFrontToBack();
}